use std::{collections::HashMap, str::FromStr, time::Duration};

use crate::{
    handling::response::{cookie_management::error_types::CookieError, CookieBuilder},
    utils::{url_utils::Uri, urlencoded::DeCodable},
};

macro_rules! update_map {
    ($map:expr, $key:expr) => {
        *$map.get_mut($key).unwrap() = true;
    };
}

use super::error_types::{ParseCookieError, ParseSameSiteError, SameSiteError};

/// Structure representing a Cookie
/// # Creating a Cookie:
/// ```
/// use http::handling::response::Cookie;
/// use http::handling::response::CookieBuilder;
///
/// let cookie = CookieBuilder::build("name", "value").finish();
/// ```
#[derive(PartialEq, Eq, PartialOrd, Ord, Debug)]
pub struct Cookie {
    /// Storage for the cookie string. Only used if this structure was derived
    /// from a string that was subsequently parsed.
    pub(crate) cookie_string: Option<String>,
    pub(crate) name: String,
    pub(crate) value: String,
    /// Whether this cookie was marked Secure.
    pub(crate) secure: bool,
    /// Whether this cookie was marked HttpOnly.
    pub(crate) http_only: bool,
    pub(crate) partitioned: bool,
    // expires: Option<Tm>,
    pub(crate) max_age: Option<Duration>,
    /// The draft `SameSite` attribute.
    pub(crate) same_site: Option<SameSite>,
    /// The cookie's domain, if any.
    pub(crate) domain: Option<String>,
    /// The cookie's path domain, if any.
    pub(crate) path: Option<Uri>,
    pub(crate) expires: Option<String>,
}

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
/// SameSite Paremeters
pub enum SameSite {
    /// Requires Secure
    None,
    Lax,
    Strict,
}

impl std::fmt::Display for SameSite {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::None => write!(f, "SameSite=None"),
            Self::Lax => write!(f, "SameSite=Lax"),
            Self::Strict => write!(f, "SameSite=Strict"),
        }
    }
}

impl FromStr for SameSite {
    type Err = ParseSameSiteError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "None" => Ok(SameSite::None),
            "Lax" => Ok(SameSite::Lax),
            "Strict" => Ok(SameSite::Strict),
            _ => Err(Self::Err {
                inner: SameSiteError::NotAValidVariant,
            }),
        }
    }
}

impl std::fmt::Display for Cookie {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let mut appendix = String::from("");
        if self.secure {
            appendix += "; Secure";
        }
        if self.http_only {
            appendix += "; HttpOnly";
        }
        if self.partitioned {
            appendix += "; Partitioned";
        }
        if let Some(max_age) = &self.max_age {
            appendix += &format!("; Max-Age={}", max_age.as_secs());
        }
        if let Some(domain) = &self.domain {
            appendix += &format!("; Domain={}", domain);
        }
        if let Some(path) = &self.path {
            appendix += &format!("; Path={}", path);
        }
        if let Some(same_site) = &self.same_site {
            appendix += &format!("; {}", same_site);
            if !self.secure && *same_site == SameSite::None {
                appendix += &format!("; Secure");
            }
        }
        if let Some(expires) = &self.expires {
            appendix += &format!("; Expires={}", expires)
        }
        write!(f, "Set-Cookie: {}={}{}", self.name, self.value, appendix)
    }
}

impl FromStr for Cookie {
    type Err = ParseCookieError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut map = HashMap::with_capacity(8);
        map.insert("Partitioned", false);
        map.insert("HttpOnly", false);
        map.insert("Secure", false);
        map.insert("SameSite", false);
        map.insert("Max-Age", false);
        map.insert("Domain", false);
        map.insert("Path", false);
        map.insert("Expires", false);
        let mut final_result = CookieBuilder::build("", "");
        let mut first = true;
        let stripped = s.strip_prefix("Set-Cookie: ").ok_or(Self::Err {
            inner: CookieError::NoSetCookieHeader,
        })?;
        for part in stripped.split(';') {
            let trimmed_part = part.trim();
            if first {
                let Some(name_val) = part.split_once('=') else {
                    return Err(Self::Err { inner: CookieError::MissingEqual  });
                };
                unsafe {
                    final_result = CookieBuilder::build(
                        &String::from_utf8_unchecked(if let Ok(name) = name_val.0.decode() {
                            name
                        } else {
                            name_val.0.into()
                        }),
                        &String::from_utf8_unchecked(if let Ok(value) = name_val.1.decode() {
                            value
                        } else {
                            name_val.1.into()
                        }),
                    );
                }
                first = false;
                continue;
            }
            if !map.get("Max-Age").unwrap() && trimmed_part.starts_with("Max-Age=") {
                final_result = final_result.max_age(Duration::from_secs(
                    trimmed_part
                        .strip_prefix("Max-Age=")
                        .unwrap()
                        .parse()
                        .map_err(|_| Self::Err {
                            inner: CookieError::InvalidMaxAge,
                        })?,
                ));
                update_map!(map, "Max-Age");
                continue;
            }
            if !map.get("Expires").unwrap() && trimmed_part.starts_with("Expires=") {
                final_result = final_result.expires(trimmed_part.strip_prefix("Expires=").unwrap());
                update_map!(map, "Expires");
                continue;
            }
            if !map.get("HttpOnly").unwrap() && trimmed_part == "HttpOnly" {
                final_result = final_result.http_only(true);
                update_map!(map, "HttpOnly");
                continue;
            }
            if !map.get("SameSite").unwrap() && trimmed_part.starts_with("SameSite=") {
                final_result = final_result.same_site(
                    trimmed_part
                        .strip_prefix("SameSite=")
                        .unwrap()
                        .parse::<SameSite>()
                        .map_err(|err| Self::Err {
                            inner: CookieError::InvilidSameSite(err.inner),
                        })?,
                );
                update_map!(map, "SameSite");
                continue;
            }
            if !map.get("Path").unwrap() && trimmed_part.starts_with("Path=") {
                final_result = final_result.path(
                    trimmed_part
                        .strip_prefix("Path=")
                        .unwrap()
                        .parse::<Uri>()
                        .map_err(|err| Self::Err {
                            inner: CookieError::PathError(err.error),
                        })?,
                );
                update_map!(map, "Path");
                continue;
            }
            if !map.get("Domain").unwrap() && trimmed_part.starts_with("Domain=") {
                final_result = final_result.domain(trimmed_part.strip_prefix("Domain=").unwrap());
                update_map!(map, "Domain");
                continue;
            }
            if !map.get("Secure").unwrap() && trimmed_part == "Secure" {
                final_result = final_result.secure(true);
                update_map!(map, "Secure");
                continue;
            }

            if !map.get("Partitioned").unwrap() && trimmed_part == "Partitioned" {
                final_result = final_result.partitioned(true);
                update_map!(map, "Partitioned");
                continue;
            }
        }
        Ok(final_result.finish())
    }
}

#[cfg(test)]
mod test {
    use std::time::Duration;

    use crate::handling::response::{Cookie, CookieBuilder};

    use super::SameSite;

    #[test]
    fn test_cookie_to_string() {
        let test_cookie1 = CookieBuilder::build("a", "cookie").finish().to_string();
        let test_cookie1_res = "Set-Cookie: a=cookie";
        let test_cookie2 = CookieBuilder::build("a", "secure_cookie")
            .secure(true)
            .finish()
            .to_string();
        let test_cookie2_res = "Set-Cookie: a=secure_cookie; Secure";
        let test_cookie3 = CookieBuilder::build("ab", "ss")
            .max_age(Duration::from_secs(24))
            .domain("codecraft.com")
            .path("/".parse().unwrap())
            .same_site(SameSite::None)
            .http_only(true)
            .partitioned(true)
            .expires("Monday")
            .finish();
        let test_cookie3_res = "Set-Cookie: ab=ss; Secure; HttpOnly; Partitioned; \
            Max-Age=24; Domain=codecraft.com; Path=/; SameSite=None; Expires=Monday";

        assert_eq!(test_cookie1_res, test_cookie1);
        assert_eq!(test_cookie2_res, test_cookie2);
        assert_eq!(test_cookie3_res, test_cookie3.to_string());
    }
    #[test]
    fn cookie_from_string() {
        let test_cookie3_res = "Set-Cookie: ab=ss; HttpOnly; Partitioned; \
            Max-Age=24; Domain=codecraft.com; Path=/; SameSite=None; Secure; Expires=Monday";
        let test_cookie3 = CookieBuilder::build("ab", "ss")
            .max_age(Duration::from_secs(24))
            .domain("codecraft.com")
            .path("/".parse().unwrap())
            .same_site(SameSite::None)
            .http_only(true)
            .partitioned(true)
            .expires("Monday")
            .finish();
        assert_eq!(test_cookie3, test_cookie3_res.parse::<Cookie>().unwrap());
    }
}
