use std::str::FromStr;

use crate::utils::{url_utils::datatypes::RawUriElement, urlencoded::UrlEncodeData};

use super::datatypes::{ParseUriError, RawUri, Uri, UriError};

impl Uri {
    pub fn new(parts: Vec<&str>) -> Self {
        Self {
            parts: parts
                .into_iter()
                .map(|part| UrlEncodeData::from_raw(part))
                .collect(),
        }
    }
}

impl RawUri {
    pub fn new(parts: Vec<&str>) -> Self {
        let mut result = Self {
            infinte_end: false,
            parts: Vec::with_capacity(parts.len()),
            raw_string: "/".to_owned() + &parts.join("/"),
        };
        for part in parts {
            if part.starts_with("<") && part.ends_with("..>") {
                result.infinte_end = true;
                break;
            }
            if part.starts_with("<") && part.ends_with(">") {
                result.parts.push(RawUriElement::Variable);
                continue;
            }
            result
                .parts
                .push(RawUriElement::Name(UrlEncodeData::from_raw(part)))
        }
        result
    }
    pub fn compare_uri(self, uri: Uri) -> bool {
        let mut iter_comp = uri.parts.iter();
        let mut counter = 0;
        for element in self.parts.iter() {
            counter += 1;
            let Some(compare_element) = iter_comp.next() else {
                return false;
            };

            if *element == RawUriElement::Variable {
                continue;
            }
            if let RawUriElement::Name(name) = element {
                if name.encoded() != compare_element.encoded() {
                    return false;
                }
            } else {
                return false;
            }
        }
        if counter > self.parts.len() && !self.infinte_end {
            return false;
        }
        true
    }
}

impl std::fmt::Display for RawUri {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.raw_string)
    }
}

impl std::fmt::Display for Uri {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let url = self
            .parts
            .iter()
            .map(|part| part.encoded())
            .collect::<Vec<_>>();
        write!(f, "/{}", url.join("/"))
    }
}

impl FromStr for Uri {
    type Err = ParseUriError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let split = s.split('/');
        let mut result = Vec::new();
        for sub in split {
            if sub.is_empty() {
                continue;
            }
            result.push(if let Ok(coded) = UrlEncodeData::from_encoded(sub) {
                coded
            } else {
                UrlEncodeData::from_raw(sub)
            });
        }
        Ok(Self { parts: result })
    }
}

impl FromStr for RawUri {
    type Err = UriError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let parts = s.split('/').collect::<Vec<&str>>();
        let mut result = Self {
            infinte_end: false,
            parts: Vec::new(),
            raw_string: parts.join("/"),
        };
        for part in parts {
            if part.is_empty() {
                continue;
            }
            if part.starts_with("<") && part.ends_with("..>") {
                result.infinte_end = true;
                break;
            }
            if part.starts_with("<") && part.ends_with(">") {
                result.parts.push(RawUriElement::Variable);
                continue;
            }
            result
                .parts
                .push(RawUriElement::Name(UrlEncodeData::from_raw(part)))
        }
        Ok(result)
    }
}

#[cfg(test)]
mod test {
    use crate::utils::{
        url_utils::{
            datatypes::{RawUri, RawUriElement},
            Uri,
        },
        urlencoded::UrlEncodeData,
    };

    #[test]
    fn uri_to_string() {
        assert_eq!("/a/%20", Uri::new(vec!["a", " "]).to_string());
    }

    #[test]
    fn uri_from_string() {
        assert_eq!(Uri::new(vec!["a", " "]), "/a/%20".parse().unwrap())
    }

    #[test]
    fn raw_uri_from_string() {
        assert_eq!(
            RawUri {
                raw_string: "/<name>/a/<f..>".to_owned(),
                infinte_end: true,
                parts: vec![
                    RawUriElement::Variable,
                    RawUriElement::Name(UrlEncodeData::from_raw("a")),
                ]
            },
            "/<name>/a/<f..>".parse().unwrap()
        );
    }

    #[test]
    fn raw_uri_to_string() {
        assert_eq!(
            "/<name>/a/<f..>",
            RawUri {
                raw_string: "/<name>/a/<f..>".to_owned(),
                infinte_end: true,
                parts: vec![
                    RawUriElement::Variable,
                    RawUriElement::Name(UrlEncodeData::from_raw("a")),
                ]
            }
            .to_string()
        )
    }
}
